from typing import Tuple, Union
from pathlib import Path

from torchvision import transforms
from torchvision.transforms import Compose
from torch import Tensor
from PIL import Image


def process_image(path: Union[Path, str], img_transforms=None) -> Tensor:
    if img_transforms is None:
        img_transforms = default_transforms()

    img = Image.open(path)
    img = img_transforms(img)
    return img


def default_transforms(size: Tuple[int, int] = (128, 128)) -> Compose:
    return transforms.Compose([
        transforms.ToTensor(),
        transforms.Resize(size),
        transforms.Normalize(
            mean=[0.485, 0.456, 0.406],
            std=[0.229, 0.224, 0.225],
            inplace=True,
        )
    ])



import os
import cv2
import numpy as np
from typing import Tuple
from os.path import join, isfile
from PIL import Image
from torch.utils.data import Dataset
from torchvision.transforms import transforms, Resize, InterpolationMode


class TestDataset(Dataset):

    def __init__(self, path_to_dataset, image_size=256, mask_size=1024,
                 mean: Tuple[float] = (0.485, 0.456, 0.406), std: Tuple[float] = (0.229, 0.224, 0.225)):
        self.path_to_dataset = path_to_dataset
        self.test_data_path = path_to_dataset + '/test'
        self.gt_data_path = path_to_dataset + '/ground_truth'

        self.image_size = image_size
        self.mask_size = mask_size
        self.mean = list(mean)
        self.std = list(std)

        self.masks, self.image_paths = self.__load_images_and_masks()
        self.len = len(self.image_paths)

        self.__build_transforms()

    def __getitem__(self, index):
        image_path = self.image_paths[index]
        img = Image.open(image_path).convert('RGB')

        original = self.resize_transform(img)
        preprocessed = self.transform(original)

        mask = self.masks[index]
        mask = np.resize(mask, (self.mask_size, self.mask_size))

        return np.array(original), preprocessed, mask

    def __len__(self):
        return self.len

    def __load_images_and_masks(self):
        masks = []
        image_paths = []

        dirs = [f for f in os.listdir(self.test_data_path)]
        dirs.sort()

        for image_dir in dirs:
            img_dir = join(self.test_data_path, image_dir)
            mask_dir = join(self.gt_data_path, image_dir)

            img_paths = [f for f in os.listdir(img_dir) if isfile(join(img_dir, f))]
            img_paths.sort()
            for path in img_paths:
                image_paths.append(join(img_dir, path))

            if image_dir == 'good':
                for _ in img_paths:
                    mask = np.zeros(shape=(self.mask_size, self.mask_size))
                    masks.append(mask)
            else:
                mask_paths = [f for f in os.listdir(mask_dir) if isfile(join(mask_dir, f))]
                mask_paths.sort()
                for path in mask_paths:
                    mask = self.__load_mask(mask_dir, path)
                    masks.append(mask)

        return masks, image_paths

    def __load_image(self, image_dir, image_name):
        img = Image.open(join(image_dir, image_name)).convert('RGB')

        return img

    def __load_mask(self, mask_dir, mask_name):
        mask = cv2.imread(join(mask_dir, mask_name), cv2.IMREAD_UNCHANGED)
        mask = cv2.resize(mask, (self.mask_size, self.mask_size))
        mask = mask / 255

        return mask

    def __build_transforms(self):
        resize_transform = Resize(size=(self.image_size, self.image_size), interpolation=InterpolationMode.BILINEAR,
                                  antialias=True)
        self.resize_transform = transforms.Compose([resize_transform])

        normalize = transforms.Normalize(mean=self.mean, std=self.std)

        self.transform = transforms.Compose([transforms.ToTensor(), normalize])



import os
from os.path import join

import numpy as np
import torch

from src.anoseg.datasets.dataset import Dataset
from src.anoseg.models.PaDiM.backbone.padim import PaDiM


class Trainer(object):

    # region init

    def __init__(self, output_dir: str, dataset: Dataset, batch_size: int = 32, num_embeddings=130,
                 backbone="resnet18", image_size: int = 512, debugging: bool = False):
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.output_dir = output_dir
        self.dataset = dataset
        self.batch_size = batch_size
        self.image_size = image_size
        self.num_embeddings = num_embeddings
        self.backbone = backbone
        self.debugging = debugging

        self.__init_models()

    def __init_models(self) -> None:
        self.padim_big = PaDiM(num_embeddings=self.num_embeddings, device=self.device, backbone=self.backbone,
                               size=(self.image_size, self.image_size))
        self.padim_medium = PaDiM(num_embeddings=self.num_embeddings, device=self.device, backbone=self.backbone,
                                  size=(self.image_size, self.image_size))
        self.padim_small = PaDiM(num_embeddings=self.num_embeddings, device=self.device, backbone=self.backbone,
                                 size=(self.image_size, self.image_size))

    # endregion

    # region public methods

    def train(self) -> None:
        if os.path.exists(os.path.join(self.output_dir, 'big', 'covs.npy')):
            self.__log("Big model already exists. Skipped training big model.")
        else:
            self.__train_big_model()
            self.__log("Finished training PaDiM model for big patches.")

        if os.path.exists(os.path.join(self.output_dir, 'medium', 'covs.npy')):
            self.__log("Medium model already exists. Skipped training medium model.")
        else:
            self.__train_medium_model()
            self.__log("Finished training PaDiM model for medium patches.")

        if os.path.exists(os.path.join(self.output_dir, 'small', 'covs.npy')):
            self.__log("Small model already exists. Skipped training small model.")
        else:
            self.__train_small_model()
            self.__log("Finished training PaDiM model for small patches.")

    # endregion

    # region private methods

    def __train_big_model(self) -> None:
        train_loader = self.dataset.get_train_dataloader(self.batch_size)
        self.padim_big.train(train_loader, epochs=3)

        N, means, covs, embedding_ids = self.padim_big.get_residuals()

        self.__save_model(N, means, covs, embedding_ids, dir_name="big")

    def __train_medium_model(self) -> None:
        patch_train_loader = self.dataset.get_medium_patches_train_dataloader(self.batch_size)
        self.padim_medium.train(patch_train_loader, epochs=3, use_patches=True)

        N, means, covs, embedding_ids = self.padim_medium.get_residuals()

        self.__save_model(N, means, covs, embedding_ids, dir_name="medium")

    def __train_small_model(self) -> None:
        patch_train_loader = self.dataset.get_small_patches_train_dataloader(self.batch_size)
        self.padim_small.train(patch_train_loader, epochs=3, use_patches=True)

        N, means, covs, embedding_ids = self.padim_small.get_residuals()

        self.__save_model(N, means, covs, embedding_ids, dir_name="small")

    def __save_model(self, N, means, covs, embedding_ids, dir_name: str) -> None:
        model_dir = join(self.output_dir, dir_name)
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)

        n_paths = join(model_dir, "n.npy")
        np.save(n_paths, N)

        means_paths = join(model_dir, "means.npy")
        np.save(means_paths, means)

        covs_paths = join(model_dir, "covs.npy")
        np.save(covs_paths, covs)

        embedding_ids_paths = join(model_dir, "embedding_ids.npy")
        np.save(embedding_ids_paths, embedding_ids)

    def __log(self, message: str) -> None:
        if self.debugging:
            print(message)

    # endregion



from typing import Tuple, Union

from numpy import ndarray as NDArray

import torch
from torch import Tensor, device as Device
from torch.utils.data import DataLoader

from src.anoseg.models.PaDiM.backbone.base import PaDiMBase
from src.anoseg.models.PaDiM.utils.distance import mahalanobis_multi, mahalanobis_sq


class PaDiM(PaDiMBase):
    """
    The PaDiM model
    """

    def __init__(
        self,
        num_embeddings: int = 100,
        device: Union[str, Device] = "cpu",
        backbone: str = "wide_resnet50",
        size: Union[None, Tuple[int, int]] = None,
    ):
        super(PaDiM, self).__init__(num_embeddings, device, backbone, size)
        self.N = 0
        self.means = torch.zeros(
            (self.num_patches, self.num_embeddings)).to(self.device)
        self.covs = torch.zeros((self.num_patches, self.num_embeddings,
                                 self.num_embeddings)).to(self.device)

    def train_one_batch(self, imgs: Tensor) -> None:
        """
        Handle only one batch, updating the internal state
        Params
        ======
            imgs: Tensor - batch tensor of size (b * c * w * h)
        """
        with torch.no_grad():
            # b * c * w * h
            embeddings = self._embed_batch(imgs.to(self.device))
            b = embeddings.size(0)
            embeddings = embeddings.reshape(
                (-1, self.num_embeddings, self.num_patches))  # b * c * (w * h)
            for i in range(self.num_patches):
                patch_embeddings = embeddings[:, :, i]  # b * c
                for j in range(b):
                    self.covs[i, :, :] += torch.outer(
                        patch_embeddings[j, :],
                        patch_embeddings[j, :])  # c * c
                self.means[i, :] += patch_embeddings.sum(dim=0)  # c
            self.N += b  # number of images

    def train(self, dataloader: DataLoader, epochs=1, use_patches=False) -> Tuple[Tensor, Tensor, Tensor]:
        """
        End-to-end training of the model
        Params
        ======
            dataloader: DataLoader - a dataset dataloader feeding images
        Returns
        =======
            means: Tensor - the computed mean vectors
            covs: Tensor - the computed covariance matrices
        """
        for e in range(epochs):
            print("Epoch: ", e)
            for imgs in dataloader:
                if not use_patches:
                    self.train_one_batch(imgs)
                else:
                    for patches in imgs:
                        self.train_one_batch(patches)

        means, covs, embedding_ids = self.get_params()
        return means, covs, embedding_ids

    def get_params(self,
                   epsilon: float = 0.01) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Computes the mean vectors and covariance matrices from the
        indermediary state
        Params
        ======
            epsilon: float - coefficient for the identity matrix
        Returns
        =======
            means: Tensor - the computed mean vectors
            covs: Tensor - the computed covariance matrices
            embedding_ids: Tensor - the embedding indices
        """
        means = self.means.detach().clone()
        covs = self.covs.detach().clone()

        identity = torch.eye(self.num_embeddings).to(self.device)
        means /= self.N
        for i in range(self.num_patches):
            covs[i, :, :] -= self.N * torch.outer(means[i, :], means[i, :])
            covs[i, :, :] /= self.N - 1  # corrected covariance
            covs[i, :, :] += epsilon * identity  # constant term

        return means, covs, self.embedding_ids

    def _get_inv_cvars(self, covs: Tensor) -> NDArray:
        inv_cvars = torch.inverse(covs)
        return inv_cvars

    def predict(self,
                new_imgs: Tensor,
                params: Tuple[Tensor, Tensor] = None,
                compare_all: bool = False) -> Tensor:
        """
        Computes the distance matrix for each image * patch
        Params
        ======
            imgs: Tensor - (b * W * H) tensor of images
            params: [(Tensor, Tensor)] - optional precomputed parameters
        Returns
        =======
            distances: Tensor - (c * b) array of distances
        """
        if params is None:
            means, covs, _ = self.get_params()
            inv_cvars = self._get_inv_cvars(covs)
        else:
            means, inv_cvars = params
        embeddings = self._embed_batch(new_imgs)
        b, c, w, h = embeddings.shape
        # not required, but need changing of testing code
        assert b == 1, f"The batch should be of size 1, got b={b}"
        embeddings = embeddings.reshape(c, w * h).permute(1, 0)

        if compare_all:
            distances = mahalanobis_multi(embeddings, means, inv_cvars)
            distances, _ = distances.min(dim=0)
        else:
            distances = mahalanobis_sq(embeddings, means, inv_cvars)
        return torch.sqrt(distances)

    def get_residuals(self) -> Tuple[int, NDArray, NDArray, NDArray]:
        """
        Get the intermediary data needed to stop the training and resume later
        Returns
        =======
            N: int - the number of images
            means: Tensor - the sums of embedding vectors
            covs: Tensor - the sums of the outer product of embedding vectors
            embedding_ids: Tensor - random dimensions used for size reduction
        """

        def detach_numpy(t: Tensor) -> NDArray:
            return t.detach().cpu().numpy()

        return (self.N, detach_numpy(self.means), detach_numpy(self.covs),
                detach_numpy(self.embedding_ids))

    @staticmethod
    def from_residuals(N: int, means: NDArray, covs: NDArray,
                       embedding_ids: NDArray, backbone: str,
                       device: Union[Device, str], img_size):
        num_embeddings, = embedding_ids.shape
        padim = PaDiM(num_embeddings=num_embeddings,
                      device=device,
                      backbone=backbone,
                      size=(img_size, img_size))
        padim.embedding_ids = torch.tensor(embedding_ids).to(device)
        padim.N = N
        padim.means = torch.tensor(means).to(device)
        padim.covs = torch.tensor(covs).to(device)

        return padim



import random
import numpy as np
from PIL import Image
from torchvision import transforms

"""
    Implementation based on https://github.com/LilitYolyan/CutPaste
    Paper: https://arxiv.org/abs/2104.04015
"""


class CutPaste(object):

    def __init__(self, mode="all"):
        self.mode = mode

    @staticmethod
    def crop_and_paste_patch(image, patch_w, patch_h, rotation=False):
        """
        Crop patch from original image and paste it randomly on the same image.

        :image: [PIL] _ original image
        :patch_w: [int] _ width of the patch
        :patch_h: [int] _ height of the patch
        :transform: [binary] _ if True use Color Jitter augmentation
        :rotation: [binary[ _ if True randomly rotates image from (-45, 45) range

        :return: augmented image
        """

        org_w, org_h = image.size
        mask = None

        patch_left, patch_top = random.randint(0, org_w - patch_w), random.randint(0, org_h - patch_h)
        patch_right, patch_bottom = patch_left + patch_w, patch_top + patch_h
        patch = image.crop((patch_left, patch_top, patch_right, patch_bottom))

        if rotation:
            random_rotate = random.uniform(*rotation)
            patch = patch.convert("RGBA").rotate(random_rotate, expand=True)
            mask = patch.split()[-1]

        # new location
        paste_left, paste_top = random.randint(0, org_w - patch_w), random.randint(0, org_h - patch_h)
        aug_image = image.copy()
        aug_image.paste(patch, (paste_left, paste_top), mask=mask)

        label_img = Image.new("RGB", image.size, (255, 255, 255))
        label_img.paste(patch, (paste_left, paste_top), mask=mask)

        label_mask = np.all(np.array(label_img) == [255, 255, 255], axis=-1)
        label = np.where(label_mask, 0, 1)

        return aug_image, label

    def cutpaste(self, image, area_ratio=(0.02, 0.15), aspect_ratio=((0.3, 1), (1, 3.3))):
        '''
        CutPaste augmentation

        :image: [PIL] - original image
        :area_ratio: [tuple] - range for area ratio for patch
        :aspect_ratio: [tuple] -  range for aspect ratio

        :return: PIL image after CutPaste transformation
        '''

        img_area = image.size[0] * image.size[1]
        patch_area = random.uniform(*area_ratio) * img_area
        patch_aspect = random.choice([random.uniform(*aspect_ratio[0]), random.uniform(*aspect_ratio[1])])
        patch_w = int(np.sqrt(patch_area * patch_aspect))
        patch_h = int(np.sqrt(patch_area / patch_aspect))
        cutpaste, mask = self.crop_and_paste_patch(image, patch_w, patch_h, rotation=False)

        return cutpaste, mask

    def cutpaste_scar(self, image, width=[2, 16], length=[10, 25], rotation=(-45, 45)):
        '''

        :image: [PIL] - original image
        :width: [list] - range for width of patch
        :length: [list] - range for length of patch
        :rotation: [tuple] - range for rotation

        :return: PIL image after CutPaste-Scare transformation
        '''
        patch_w, patch_h = random.randint(*width), random.randint(*length)
        cutpaste_scar, mask = self.crop_and_paste_patch(image, patch_w, patch_h, rotation=rotation)

        return cutpaste_scar, mask

    def __call__(self, image):
        '''
        :image: [PIL] - original image
        :return: returns original image and randomly chosen transformation
        '''

        if self.mode == "all":
            aug = random.choice([self.cutpaste, self.cutpaste_scar])
            abnormal_img, label = aug(image)
        elif self.mode == "scar":
            abnormal_img, label = self.cutpaste_scar(image)
        elif self.mode == "patch":
            abnormal_img, label = self.cutpaste(image)
        else:
            print("Unknown cutpaste mode")
            return

        mask_normal = np.zeros(image.size, dtype=np.float32)

        return np.array(image), np.array(abnormal_img), mask_normal, label



import os
import torchvision.transforms.functional as TF
import torch
import torch.nn.functional as F
import numpy as np
import torchvision
from typing import List
from PIL import Image
from torch import Tensor
from torchvision.transforms import transforms
from typing_extensions import Literal

from src.anoseg.models.DFC.backbone.vgg19 import VGG19
from src.anoseg.models.DFC.backbone.vgg19_s import VGG19_S
from src.anoseg.models.utils import BaseTester


class Tester(BaseTester):

    # region init

    _DATASET_TYPES = Literal["textures", "objects"]
    _CNN_LAYERS_TEXTURES = ("relu4_1", "relu4_2", "relu4_3", "relu4_4")
    _CNN_LAYERS_OBJECTS = ("relu4_3", "relu4_4", "relu5_1", "relu5_2")

    def __init__(self, model_path: str, dataset_type: _DATASET_TYPES, pretrained_weights_dir: str = None,
                 debugging: bool = False,
                 image_size: int = 256, mask_size: int = 1024, use_self_ensembling: bool = False,
                 rot_90: bool = False, rot_180: bool = False, rot_270: bool = False, h_flip: bool = False,
                 h_flip_rot_90: bool = False, h_flip_rot_180: bool = False, h_flip_rot_270: bool = False,
                 integration_limit: float = 0.3):
        super().__init__(model_path=model_path, debugging=debugging, image_size=image_size, mask_size=mask_size,
                         use_self_ensembling=use_self_ensembling, rot_90=rot_90, rot_180=rot_180, rot_270=rot_270,
                         h_flip=h_flip, h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                         h_flip_rot_270=h_flip_rot_270, integration_limit=integration_limit)

        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.dataset_type = dataset_type
        self.pretrained_weights_dir = pretrained_weights_dir

        self.__init_feat_layers()
        self.__load_model()

    def __init_feat_layers(self) -> None:
        if self.dataset_type == "objects":
            self.feat_layers = self._CNN_LAYERS_OBJECTS
            self._log_message("Using object layers.")
        elif self.dataset_type == "textures":
            self.feat_layers = self._CNN_LAYERS_TEXTURES
            self._log_message("Using texture layers.")
        else:
            raise Exception("Unknown dataset type. Valid types: ['textures', 'objects']")

    def __load_model(self) -> None:
        # pretrained feature extraction net
        self.feature_extraction = VGG19(pretrain=True, gradient=False, pool='avg',
                                        pretrained_weights_dir=self.pretrained_weights_dir).to(self.device)
        # trained feature estimation net
        self.feature_matching = VGG19_S(pretrain=False, gradient=True, pool='avg').to(self.device)
        self.feature_matching.load_state_dict(torch.load(os.path.join(self.model_path, 'match.pt'),
                                                         map_location=torch.device(self.device)))

        self.feature_extraction.eval()
        self.feature_matching.eval()

    # endregion

    # region implement abstract methods

    def score(self, img_input) -> np.array:  # returns score with shape (1024, 1024)
        img = img_input.to(self.device)
        with torch.no_grad():
            surrogate_label = self.feature_extraction(img, self.feat_layers)
            prediction = self.feature_matching(img, self.feat_layers)

        anomaly_map = 0
        for feat_layer in self.feat_layers:
            anomaly_map += F.interpolate(
                torch.pow(surrogate_label[feat_layer] - prediction[feat_layer], 2).mean(dim=1, keepdim=True),
                size=(self.mask_size, self.mask_size), mode="bilinear", align_corners=True)

        scores = anomaly_map.data.cpu().numpy().squeeze()

        return scores

    def score_with_augmentation(self, img_input) -> np.array:
        score_list = self._get_self_ensembling_scores(img_input)
        final_score = self._combine_scores(score_list)

        return final_score

    def preprocess_img(self, image_path: str, mean: List[float], std: List[float]) -> Tensor:
        original = Image.open(image_path).convert('RGB')

        normalize = transforms.Normalize(mean=mean, std=std)
        resize = torchvision.transforms.Resize(size=self.image_size, interpolation=TF.InterpolationMode.BILINEAR,
                                               antialias=True)

        transform = transforms.Compose([transforms.ToTensor(), normalize, resize])
        preprocessed = transform(original)

        preprocessed = preprocessed[None, :]

        return preprocessed

    # endregion



import torchvision.transforms.functional as TF
import numpy as np
import torch
from torch import Tensor
from torch.utils.data import DataLoader
from typing import List, Tuple, Dict
from abc import abstractmethod

import src.anoseg.evaluation.eval as evaluation
from src.anoseg.datasets.dataset import Dataset
from src.anoseg.utils import visualization


class BaseTester(object):

    # region init

    def __init__(self, model_path: str, debugging: bool = False,
                 image_size: int = 256, mask_size: int = 1024, use_self_ensembling: bool = False,
                 rot_90: bool = False, rot_180: bool = False, rot_270: bool = False, h_flip: bool = False,
                 h_flip_rot_90: bool = False, h_flip_rot_180: bool = False, h_flip_rot_270: bool = False,
                 integration_limit: float = 0.3):
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.model_path = model_path
        self.debugging = debugging
        self.image_size = image_size
        self.mask_size = mask_size

        self.use_self_ensembling = use_self_ensembling
        self.rot_90 = rot_90
        self.rot_180 = rot_180
        self.rot_270 = rot_270
        self.h_flip = h_flip
        self.h_flip_rot_90 = h_flip_rot_90
        self.h_flip_rot_180 = h_flip_rot_180
        self.h_flip_rot_270 = h_flip_rot_270

        self.integration_limit = integration_limit

    # endregion

    # region public methods

    def evaluate(self, dataset: Dataset) -> Dict[str, float]:
        scores, masks = self.__predict_images(dataset)

        return evaluation.get_metrics(scores=scores, masks=masks, debugging=self.debugging)

    def predict(self, image_path: str,
                mean: Tuple[float] = (0.485, 0.456, 0.406), std: Tuple[float] = (0.229, 0.224, 0.225)):
        preprocessed = self.preprocess_img(image_path=image_path,
                                           mean=list(mean),
                                           std=list(std))

        if self.use_self_ensembling:
            score = self.score_with_augmentation(preprocessed)
        else:
            score = self.score(preprocessed)

        binary_score = evaluation.get_binary_score(score)

        return score, binary_score

    def display_predictions(self, dataset: Dataset) -> None:
        test_dataloader: DataLoader = dataset.get_test_dataloader()

        for original, preprocessed, mask in test_dataloader:
            if self.use_self_ensembling:
                score = self.score_with_augmentation(preprocessed)
            else:
                score = self.score(preprocessed)

            original = original.squeeze()
            mask = mask.squeeze()
            binary_score = evaluation.get_binary_score(score)
            visualization.display_images(img_list=[original, mask, score, binary_score],
                                         titles=['original', 'ground_truth', 'score', 'binary_score'],
                                         cols=3)

    # endregion

    # region abstract methods

    @abstractmethod
    def score(self, img_input) -> np.array:
        pass

    @abstractmethod
    def score_with_augmentation(self, img_input) -> np.array:
        pass

    @abstractmethod
    def preprocess_img(self, image_path: str, mean: List[float], std: List[float]) -> Tensor:
        pass

    # endregion

    # region protected methods

    def _log_message(self, message):
        if self.debugging:
            print(message)

    def _get_self_ensembling_scores(self, img_input) -> List[np.array]:
        score = self.score(img_input)
        score_list = [score]

        if self.rot_90:
            rotated_90 = TF.rotate(img_input, -90)
            rotated_90_score = self.score(rotated_90)
            rotated_90_score = np.rot90(rotated_90_score)
            score_list.append(rotated_90_score)
        if self.rot_180:
            rotated_180 = TF.rotate(img_input, -180)
            rotated_180_score = self.score(rotated_180)
            rotated_180_score = np.rot90(rotated_180_score, k=2)
            score_list.append(rotated_180_score)
        if self.rot_270:
            rotated_270 = TF.rotate(img_input, -270)
            rotated_270_score = self.score(rotated_270)
            rotated_270_score = np.rot90(rotated_270_score, k=3)
            score_list.append(rotated_270_score)
        if self.h_flip:
            horizontal_flip = torch.flip(img_input, dims=[3])
            horizontal_flip_score = self.score(horizontal_flip)
            horizontal_flip_score = np.fliplr(horizontal_flip_score)
            score_list.append(horizontal_flip_score)
        if self.h_flip_rot_90:
            flipped_rotated_90 = TF.rotate(torch.flip(img_input, dims=[3]), -90)
            flipped_rotated_90_score = self.score(flipped_rotated_90)
            flipped_rotated_90_score = np.fliplr(np.rot90(flipped_rotated_90_score))
            score_list.append(flipped_rotated_90_score)
        if self.h_flip_rot_180:
            flipped_rotated_180 = TF.rotate(torch.flip(img_input, dims=[3]), -180)
            flipped_rotated_180_score = self.score(flipped_rotated_180)
            flipped_rotated_180_score = np.fliplr(np.rot90(flipped_rotated_180_score, k=2))
            score_list.append(flipped_rotated_180_score)
        if self.h_flip_rot_270:
            flipped_rotated_270 = TF.rotate(torch.flip(img_input, dims=[3]), -270)
            flipped_rotated_270_score = self.score(flipped_rotated_270)
            flipped_rotated_270_score = np.fliplr(np.rot90(flipped_rotated_270_score, k=3))
            score_list.append(flipped_rotated_270_score)

        return score_list

    def _combine_scores(self, score_list):
        res = np.mean(score_list, axis=0)

        return res

    # endregion

    # region private methods

    def __predict_images(self, dataset: Dataset) -> Tuple[List[np.array], List[np.array]]:
        test_dataloader: DataLoader = dataset.get_test_dataloader()

        number_of_paths = len(test_dataloader.dataset)
        self._log_message(f"Testing {number_of_paths} images.")
        count = 1

        scores = []
        masks = []

        for _, preprocessed, mask in test_dataloader:
            if count % 10 == 0:
                self._log_message("Predicting img {}/{}".format(count, number_of_paths))
            count += 1

            if self.use_self_ensembling:
                score = self.score_with_augmentation(preprocessed)
            else:
                score = self.score(preprocessed)

            scores.append(score)
            masks.append(mask.squeeze().numpy())

        return scores, masks

    # endregion



import torch
import torch.nn as nn
from src.anoseg.models.DFC.backbone.vgg import vgg19

"""
    Code based on https://github.com/YoungGod/DFC
    Paper: https://doi.org/10.1016/j.patcog.2022.108874
"""


# Res blocks
class ResnetBlock(nn.Module):
    def __init__(self, dim, dilation=2, padding='reflect'):
        super(ResnetBlock, self).__init__()
        self.conv_block = nn.Sequential(
            nn.ReflectionPad2d(dilation) if padding == 'reflect' else nn.ZeroPad2d(dilation),
            nn.Conv2d(in_channels=dim, out_channels=dim, kernel_size=3, padding=0, dilation=dilation),
            # models.InstanceNorm2d(dim, track_running_stats=False),
            nn.ReLU(True),

            nn.ReflectionPad2d(1) if padding == 'reflect' else nn.ZeroPad2d(1),
            nn.Conv2d(in_channels=dim, out_channels=dim, kernel_size=3, padding=0, dilation=1),
            # models.InstanceNorm2d(dim, track_running_stats=False),
        )

    def forward(self, x):
        out = x + self.conv_block(x)  # ReLU or not?
        # print(out.size())   ##########################################
        # Remove ReLU at the end of the residual block
        # http://torch.ch/blog/2016/02/04/resnets.html

        return out


# current Estimator is using 4 residual blocks
class MCRL(nn.Module):
    def __init__(self, in_channels=256, residual_blocks=2, dilation=2):
        super(MCRL, self).__init__()

        # deformable branch (data dependent context)
        blocks = []
        for _ in range(residual_blocks):
            block = ResnetBlock(in_channels, dilation=2, padding='reflect')
            # block = ResnetBlock_BDCM(in_channels, dilation=dilation)
            blocks.append(block)
        self.middle_res1 = nn.Sequential(*blocks)

        # dilated branch (fixed context)
        blocks = []
        for _ in range(residual_blocks):
            block = ResnetBlock(in_channels, dilation=4, padding='reflect')
            blocks.append(block)
        self.middle_res2 = nn.Sequential(*blocks)

        # dilated branch (fixed context)
        blocks = []
        for _ in range(residual_blocks):
            block = ResnetBlock(in_channels, dilation=8, padding='reflect')
            blocks.append(block)
        self.middle_res3 = nn.Sequential(*blocks)

    def forward(self, x):
        x1 = self.middle_res1(x)
        x2 = self.middle_res2(x)
        x3 = self.middle_res3(x)
        out = x + x1 + x2 + x3
        return out


class VGG19_S(torch.nn.Module):
    """
    VGG19: [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M']
    level1: 64*2=128; level2: 128*2=256; level3: 256*4=1024; level4: 512*4=2048; level5: 512*4=2048
    Total dimension: 128 + 256 + 1024 + 2048 + 2048 = 5504
    """

    def __init__(self, pretrain=False, gradient=False, pool='avg'):
        super(VGG19_S, self).__init__()
        features = vgg19(pretrained=pretrain, progress=False).features  # feature layers
        """ vgg.features
        Sequential(
          (0): Conv2d(3, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (1): ReLU(inplace)                                                        # self.relu1_1
          (2): Conv2d(64, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (3): ReLU(inplace)                                                        # self.relu1_2
          
          (4): MaxPool2d(kernel_size=2, stride=2, padding=0, dilation=1, ceil_mode=False)
          (5): Conv2d(64, 128, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (6): ReLU(inplace)
          (7): Conv2d(128, 128, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (8): ReLU(inplace)  
              
          (9): MaxPool2d(kernel_size=2, stride=2, padding=0, dilation=1, ceil_mode=False)
          (10): Conv2d(128, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (11): ReLU(inplace)
          (12): Conv2d(256, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (13): ReLU(inplace)
          (14): Conv2d(256, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (15): ReLU(inplace)
          (16): Conv2d(256, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (17): ReLU(inplace)
          
          (18): MaxPool2d(kernel_size=2, stride=2, padding=0, dilation=1, ceil_mode=False)
          (19): Conv2d(256, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (20): ReLU(inplace)
          (21): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (22): ReLU(inplace)
          (23): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (24): ReLU(inplace)
          (25): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (26): ReLU(inplace)
          
          (27): MaxPool2d(kernel_size=2, stride=2, padding=0, dilation=1, ceil_mode=False)
          (28): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (29): ReLU(inplace)
          (30): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (31): ReLU(inplace)
          (32): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (33): ReLU(inplace)
          (34): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (35): ReLU(inplace)
          
          (36): MaxPool2d(kernel_size=2, stride=2, padding=0, dilation=1, ceil_mode=False)
        )
        """
        # hierarchy 1 (level 1)
        self.conv1_1 = features[0]
        self.relu1_1 = features[1]
        self.conv1_2 = features[2]
        self.relu1_2 = features[3]

        # hierarchy 2 (level 2)
        self.pool1 = features[4]
        self.conv2_1 = features[5]
        self.relu2_1 = features[6]
        self.conv2_2 = features[7]
        self.relu2_2 = features[8]

        # replace all convs in level 3 with deformable conv v2
        # hierarchy 3 (level 3)    
        self.pool2 = features[9]
        self.conv3_1 = features[10]
        self.relu3_1 = features[11]
        self.conv3_2 = features[12]
        self.relu3_2 = features[13]
        self.conv3_3 = features[14]
        self.relu3_3 = features[15]
        self.conv3_4 = features[16]
        self.relu3_4 = features[17]

        # hierarchy 4 (level 4)
        self.pool3 = features[18]
        self.conv4_1 = features[19]
        self.relu4_1 = features[20]
        self.conv4_2 = features[21]
        self.relu4_2 = features[22]
        self.conv4_3 = features[23]
        self.relu4_3 = features[24]
        self.conv4_4 = features[25]
        self.relu4_4 = features[26]

        # after pool3
        self.resnet = MCRL(in_channels=self.conv3_4.in_channels, residual_blocks=4, dilation=2)

        # self.resnet = DilatedDenseResidualBlock(in_channels=self.conv3_4.in_channels, dilated_dense_blocks=2, growthrate=32, num_layer=4, dilation=True)
        # # init as zero
        # def weight_init(m):
        #     """
        #     ref: https://blog.csdn.net/qq_36338754/article/details/97756378
        #     """
        #     if isinstance(m, models.Conv2d) or isinstance(m, models.Linear):
        #         if m.weight is not None:
        #             m.weight.data.zero_()
        #         if m.bias is not None:
        #             m.bias.data.zero_()
        # self.resnet.apply(weight_init)    

        # hierarchy 5 (level 5)
        self.pool4 = features[27]
        self.conv5_1 = features[28]
        self.relu5_1 = features[29]
        self.conv5_2 = features[30]
        self.relu5_2 = features[31]
        self.conv5_3 = features[32]
        self.relu5_3 = features[33]
        self.conv5_4 = features[34]
        self.relu5_4 = features[35]

        self.pool5 = features[36]

        # don't need the gradients, just want the features
        if not gradient:
            for param in self.parameters():
                param.requires_grad = False

        # reflection padding
        self.pad = nn.ReflectionPad2d(padding=1)

        # avg pooling
        if pool == 'avg':
            avg_pool = nn.AvgPool2d(kernel_size=2, stride=2)
            # self.pool3 = self.pool4 = self.pool5 = avg_pool
            self.pool1 = self.pool2 = self.pool3 = self.pool4 = self.pool5 = avg_pool

    def forward(self, x, feature_layers):
        # level 1    pad before conv
        x = self.pad(x)
        conv1_1 = self.conv1_1(x)
        relu1_1 = self.relu1_1(conv1_1)
        conv1_2 = self.conv1_2(self.pad(relu1_1))
        relu1_2 = self.relu1_2(conv1_2)
        pool1 = self.pool1(relu1_2)

        # level 2
        pool1 = self.pad(pool1)
        conv2_1 = self.conv2_1(pool1)
        relu2_1 = self.relu2_1(conv2_1)
        conv2_2 = self.conv2_2(self.pad(relu2_1))
        relu2_2 = self.relu2_2(conv2_2)
        pool2 = self.pool2(relu2_2)

        # level 3
        pool2 = self.pad(pool2)
        conv3_1 = self.conv3_1(pool2)
        relu3_1 = self.relu3_1(conv3_1)
        conv3_2 = self.conv3_2(self.pad(relu3_1))
        relu3_2 = self.relu3_2(conv3_2)
        conv3_3 = self.conv3_3(self.pad(relu3_2))
        # conv3_3 = self.conv3_3(relu3_2)
        relu3_3 = self.relu3_3(conv3_3)
        conv3_4 = self.conv3_4(self.pad(relu3_3))
        relu3_4 = self.relu3_4(conv3_4)
        pool3 = self.pool3(relu3_4)

        # level 4
        pool3 = self.resnet(pool3)  # resblock
        pool3 = self.pad(pool3)
        conv4_1 = self.conv4_1(pool3)
        relu4_1 = self.relu4_1(conv4_1)
        conv4_2 = self.conv4_2(self.pad(relu4_1))
        relu4_2 = self.relu4_2(conv4_2)
        conv4_3 = self.conv4_3(self.pad(relu4_2))
        relu4_3 = self.relu4_3(conv4_3)
        conv4_4 = self.conv4_4(self.pad(relu4_3))
        relu4_4 = self.relu4_4(conv4_4)
        pool4 = self.pool4(relu4_4)

        # level 5
        pool4 = self.pad(pool4)
        conv5_1 = self.conv5_1(pool4)
        relu5_1 = self.relu5_1(conv5_1)
        conv5_2 = self.conv5_2(self.pad(relu5_1))
        relu5_2 = self.relu5_2(conv5_2)
        # conv5_3 = self.conv5_3(self.pad(relu5_2))
        # relu5_3 = self.relu5_3(conv5_3)
        # conv5_4 = self.conv5_4(self.pad(relu5_3))
        # relu5_4 = self.relu5_4(conv5_4)
        # pool5 = self.pool5(relu5_4)

        out = {
            'relu1_1': relu1_1,
            'relu1_2': relu1_2,

            'relu2_1': relu2_1,
            'relu2_2': relu2_2,

            'relu3_1': relu3_1,
            'relu3_2': relu3_2,
            'relu3_3': relu3_3,
            'relu3_4': relu3_4,

            'relu4_1': relu4_1,
            'relu4_2': relu4_2,
            'relu4_3': relu4_3,
            'relu4_4': relu4_4,

            'relu5_1': relu5_1,
            'relu5_2': relu5_2,
            # 'relu5_3': relu5_3,
            # 'relu5_4': relu5_4,
        }
        return dict((key, value) for key, value in out.items() if key in feature_layers)


if __name__ == '__main__':
    x = torch.randn((1, 3, 256, 256), dtype=torch.float32)
    model = VGG19_S()
    y = model(x, feature_layers=("relu1_1", "relu1_2", "relu2_1", "relu2_2",
                                 "relu3_1", "relu3_2", "relu3_3", "relu3_4",
                                 "relu4_1", "relu4_2", "relu4_3", "relu4_4",
                                 "relu5_1", "relu5_2"))  # , "relu5_3", "relu5_4"
    for key, value in y.items():
        print(key, value.shape)
    print(model.conv3_1.in_channels, model.conv3_4.out_channels)
    print(model.conv4_1.in_channels, model.conv4_2.out_channels)
    print(model.conv4_1.kernel_size, model.conv4_2.out_channels)



import os

import numpy as np
from sklearn.metrics import auc
from sklearn.metrics import roc_auc_score, average_precision_score
from skimage import measure
import pandas as pd


def eval_dfc_metrics(masks, scores, max_step=200, expect_fpr=0.3):
    # as array
    masks = np.array(masks)
    scores = np.array(scores)

    # binary masks
    masks[masks <= 0.5] = 0
    masks[masks > 0.5] = 1
    masks = masks.astype(np.bool)

    # auc score (image level) for detection
    labels = masks.any(axis=1).any(axis=1)
    # preds = scores.mean(1).mean(1)
    preds = scores.max(1).max(1)  # for detection max score or mean score?
    det_auc_score = roc_auc_score(labels, preds)
    det_pr_score = average_precision_score(labels, preds)

    # auc score (per pixel level) for segmentation
    seg_auc_score = roc_auc_score(masks.ravel(), scores.ravel())
    seg_pr_score = average_precision_score(masks.ravel(), scores.ravel())
    # metrics over all data
    print(f"Det AUC: {det_auc_score:.4f}, Seg AUC: {seg_auc_score:.4f}")
    print(f"Det PR: {det_pr_score:.4f}, Seg PR: {seg_pr_score:.4f}")

    # per region overlap and per image iou
    max_th = scores.max()
    min_th = scores.min()
    delta = (max_th - min_th) / max_step

    ious_mean = []
    ious_std = []
    pros_mean = []
    pros_std = []
    threds = []
    fprs = []
    binary_score_maps = np.zeros_like(scores, dtype=np.bool)
    for step in range(max_step):
        if step % 100 == 0:
            print('Step: ', step)

        thred = max_th - step * delta
        # segmentation
        binary_score_maps[scores <= thred] = 0
        binary_score_maps[scores > thred] = 1

        pro = []  # per region overlap
        iou = []  # per image iou
        # pro: find each connected gt region, compute the overlapped pixels between the gt region and predicted region
        # iou: for each image, compute the ratio, i.e. intersection/union between the gt and predicted binary map
        for i in range(len(binary_score_maps)):  # for i th image
            # pro (per region level)
            label_map = measure.label(masks[i], connectivity=2)
            props = measure.regionprops(label_map)
            for prop in props:
                x_min, y_min, x_max, y_max = prop.bbox  # find the bounding box of an anomaly region
                cropped_pred_label = binary_score_maps[i][x_min:x_max, y_min:y_max]
                # cropped_mask = masks[i][x_min:x_max, y_min:y_max]    # bug
                cropped_mask = prop.filled_image  # corrected!
                intersection = np.logical_and(cropped_pred_label, cropped_mask).astype(np.float32).sum()
                pro.append(intersection / prop.area)
            # iou (per image level)
            intersection = np.logical_and(binary_score_maps[i], masks[i]).astype(np.float32).sum()
            union = np.logical_or(binary_score_maps[i], masks[i]).astype(np.float32).sum()
            if masks[i].any() > 0:  # when the gt have no anomaly pixels, skip it
                iou.append(intersection / union)
        # against steps and average metrics on the testing data
        ious_mean.append(np.array(iou).mean())
        # print("per image mean iou:", np.array(iou).mean())
        ious_std.append(np.array(iou).std())
        pros_mean.append(np.array(pro).mean())
        pros_std.append(np.array(pro).std())
        # fpr for pro-auc
        masks_neg = ~masks
        fpr = np.logical_and(masks_neg, binary_score_maps).sum() / masks_neg.sum()
        fprs.append(fpr)
        threds.append(thred)

    # as array
    threds = np.array(threds)
    pros_mean = np.array(pros_mean)
    pros_std = np.array(pros_std)
    fprs = np.array(fprs)

    ious_mean = np.array(ious_mean)
    ious_std = np.array(ious_std)

    # save results
    data = np.vstack([threds, fprs, pros_mean, pros_std, ious_mean, ious_std])
    df_metrics = pd.DataFrame(data=data.T, columns=['thred', 'fpr',
                                                    'pros_mean', 'pros_std',
                                                    'ious_mean', 'ious_std'])

    # best per image iou
    best_miou = ious_mean.max()
    print(f"Best IOU: {best_miou:.4f}")

    # default 30% fpr vs pro, pro_auc
    idx = fprs <= expect_fpr  # find the indexs of fprs that is less than expect_fpr (default 0.3)
    fprs_selected = fprs[idx]
    fprs_selected = rescale(fprs_selected)  # rescale fpr [0,0.3] -> [0, 1]

    pros_mean_selected = pros_mean[idx]  # no scale  (correct?)

    pro_auc_score = auc(fprs_selected, pros_mean_selected)
    print("pro auc ({}% FPR):".format(int(expect_fpr * 100)), pro_auc_score)

    return f"{det_pr_score:.5f},{det_auc_score:.5f},{seg_pr_score:.5f},{seg_auc_score:.5f},{pro_auc_score:.5f},{best_miou:.5f}"


def rescale(x):
    return (x - x.min()) / (x.max() - x.min())



import numpy as np

from src.anoseg.evaluation import iou
from src.anoseg.evaluation import dfc_metric, pro_metric, util
from sklearn.metrics import roc_auc_score

"""
    Code is based on the official evaluation code for mvtec datasets 
    https://www.mvtec.com/company/research/datasets/mvtec-ad
    
    Scientific Papers:
        https://link.springer.com/content/pdf/10.1007/s11263-020-01400-4.pdf
        https://openaccess.thecvf.com/content_CVPR_2019/papers/Bergmann_MVTec_AD_--_A_Comprehensive_Real-World_Dataset_for_Unsupervised_Anomaly_CVPR_2019_paper.pdf
"""


def get_metrics(scores, masks, binary_scores=None, debugging: bool = False):
    if binary_scores is None:
        binary_scores = calculate_binary_scores(scores)

    image_level_roc, pixel_level_roc = calculate_au_roc(ground_truth=masks,
                                                        predictions=scores)
    au_pro, pro_curve = calculate_au_pro(ground_truth=masks, predictions=scores,
                                         integration_limit=0.3)
    avg_iou = calculate_avg_iou(ground_truth=masks, binary_scores=binary_scores)

    if debugging:
        print("Image level ROC-AUC: ", image_level_roc)
        print("Pixel level ROC-AUC: ", pixel_level_roc)
        print("PRO-AUC: ", au_pro)
        print("IoU: ", avg_iou)

        print(f"({round(pixel_level_roc * 100, 1)}, {round(au_pro * 100, 1)}, {round(avg_iou * 100, 1)})")

    return {"Pixel-ROC-AUC": pixel_level_roc,
            "Image-ROC-AUC": image_level_roc,
            "PRO-AUC": au_pro,
            "IoU": iou}


def calculate_binary_scores(scores):
    binary_scores = []
    for score in scores:
        bin_score = get_binary_score(score)
        binary_scores.append(bin_score)

    return binary_scores


def get_binary_score(score):
    if not np.all(score == 0):
        score = (score - np.min(score)) / (np.max(score) - np.min(score))
    binary_score = np.zeros_like(score)
    binary_score[score <= 0.5] = 0
    binary_score[score > 0.5] = 1

    return binary_score


def calculate_dfc_metrics(ground_truth, predictions):
    dfc_metric.eval_dfc_metrics(ground_truth, predictions)


def calculate_au_roc(ground_truth, predictions):
    # Derive binary labels for each input image:
    # (0 = anomaly free, 1 = anomalous).
    binary_labels = [int(np.any(x > 0)) for x in ground_truth]

    # official MvTec eval code does the same calculations as sklearn.metrics.roc_auc_score but is slower
    """# Compute the classification ROC curve.
    fprs, tprs, optimal_threshold, highest_accuracy = roc_metric.compute_classification_roc(
        anomaly_maps=predictions,
        scoring_function=np.max,
        ground_truth_labels=binary_labels)

    roc_curve = fprs, tprs

    # Compute the area under the classification ROC curve.
    au_roc = util.calculate_auc(roc_curve[0], roc_curve[1])"""

    image_level_roc = roc_auc_score(binary_labels, np.asarray(predictions).max(1).max(1))
    pixel_level_roc = roc_auc_score(np.asarray(ground_truth).astype(np.bool).ravel(), np.asarray(predictions).ravel())

    return image_level_roc, pixel_level_roc


def calculate_au_pro(ground_truth, predictions, integration_limit):
    pro_curve = pro_metric.compute_pro(
        anomaly_maps=predictions,
        ground_truth_maps=ground_truth)

    # Compute the area under the PRO curve.
    au_pro = util.calculate_auc(
        pro_curve[0], pro_curve[1], x_max=integration_limit)
    au_pro /= integration_limit

    return au_pro, pro_curve


"""
    returns TP_rate, FP_rate, FN_rate, TP_rate, Accuracy
    
    0 = normal
    1 = abnormal
    
        |   GT  |   Prediction
    ----------------------------
    TP  |   0   |       0
    FP  |   1   |       0
    TN  |   1   |       1
    FN  |   0   |       1
"""


def calculate_rates(ground_truth, prediction_labels):
    gt_labels = [int(np.any(x > 0)) for x in ground_truth]

    n_tp = 0
    n_fp = 0
    n_fn = 0
    n_tn = 0

    for gt, pred in zip(gt_labels, prediction_labels):
        if gt == 0 and pred == 0:
            n_tp += 1
        if gt == 0 and pred == 1:
            n_fn += 1
        if gt == 1 and pred == 0:
            n_fp += 1
        if gt == 1 and pred == 1:
            n_tn += 1

    accuracy = (n_tp + n_tn) / len(prediction_labels)
    return n_tp / len(prediction_labels), n_fp / len(prediction_labels), n_fn / len(prediction_labels), n_tn / len(
        prediction_labels), accuracy


def calculate_avg_iou(ground_truth, binary_scores):
    return iou.calculate_avg_iou(ground_truth, binary_scores)



import os
import numpy as np
import torch
from os.path import join
from torchvision.models import wide_resnet50_2, Wide_ResNet50_2_Weights

from src.anoseg.datasets.dataset import Dataset


"""
    Implementation of SPADE: Sub-Image Anomaly Detection with Deep Pyramid Correspondences
    Code based on: https://github.com/byungjae89/SPADE-pytorch/tree/master
    Paper: https://arxiv.org/abs/2005.02357
"""


class Trainer(object):

    # region init

    def __init__(self, output_dir: str, dataset: Dataset, debugging: bool,
                 batch_size: int = 32, image_size: int = 256):
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.output_dir = output_dir
        self.dataset = dataset
        self.debugging = debugging
        self.batch_size = batch_size
        self.image_size = image_size

        self.__init_model()

    def __init_model(self) -> None:
        self.model = wide_resnet50_2(weights=Wide_ResNet50_2_Weights.IMAGENET1K_V1, progress=True)
        self.model.to(self.device)
        self.model.eval()

    # endregion

    # region public methods

    def train(self) -> None:
        train_loader = self.dataset.get_train_dataloader(self.batch_size)

        outputs = []

        def hook(module, input, output):
            outputs.append(output)

        self.model.layer1[-1].register_forward_hook(hook)
        self.model.layer2[-1].register_forward_hook(hook)
        self.model.layer3[-1].register_forward_hook(hook)
        self.model.avgpool.register_forward_hook(hook)

        train_outputs = [[], [], [], []]

        for img in train_loader:
            img = img.to(self.device)
            # model prediction
            with torch.no_grad():
                pred = self.model(img)
                del pred
                del img
            # get intermediate layer outputs
            for i in range(len(train_outputs)):
                train_outputs[i].append(outputs[i])
            # train_outputs = [lst + [outputs[i]] for i, lst in enumerate(train_outputs)]

            # initialize hook outputs
            outputs = []
        for i in range(len(train_outputs)):
            train_outputs[i] = np.array(torch.cat(train_outputs[i], 0).cpu())
        # train_outputs = [np.array(torch.stack(lst, dim=0).cpu()) for lst in train_outputs]

        self.__save_model(train_outputs=train_outputs)

    # endregion

    # region private methods

    def __save_model(self, train_outputs) -> None:
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

        l1_path = join(self.output_dir, "layer_1.npy")
        np.save(l1_path, train_outputs[0])

        l2_path = join(self.output_dir, "layer_2.npy")
        np.save(l2_path, train_outputs[1])

        l3_path = join(self.output_dir, "layer_3.npy")
        np.save(l3_path, train_outputs[2])

        pool_path = join(self.output_dir, "avgpool.npy")
        np.save(pool_path, train_outputs[3])

    # endregion



import torch
import torch.nn as nn
from src.anoseg.models.DFC.backbone.vgg import vgg19

"""
    Code based on https://github.com/YoungGod/DFC
    Paper: https://doi.org/10.1016/j.patcog.2022.108874
"""


class VGG19(torch.nn.Module):
    """
    VGG19: [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M']
    level1: 64*2=128; level2: 128*2=256; level3: 256*4=1024; level4: 512*4=2048; level5: 512*4=2048
    Total dimension: 128 + 256 + 1024 + 2048 + 2048 = 5504
    """

    def __init__(self, pretrain=False, gradient=False, pool='avg', pretrained_weights_dir=None):
        super(VGG19, self).__init__()
        features = vgg19(pretrained=pretrain, pretrained_weights_dir=pretrained_weights_dir).features  # feature layers
        """ vgg.features
        Sequential(
          (0): Conv2d(3, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (1): ReLU(inplace)                                                        # self.relu1_1
          (2): Conv2d(64, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (3): ReLU(inplace)                                                        # self.relu1_2
          
          (4): MaxPool2d(kernel_size=2, stride=2, padding=0, dilation=1, ceil_mode=False)
          (5): Conv2d(64, 128, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (6): ReLU(inplace)
          (7): Conv2d(128, 128, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (8): ReLU(inplace)  
              
          (9): MaxPool2d(kernel_size=2, stride=2, padding=0, dilation=1, ceil_mode=False)
          (10): Conv2d(128, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (11): ReLU(inplace)
          (12): Conv2d(256, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (13): ReLU(inplace)
          (14): Conv2d(256, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (15): ReLU(inplace)
          (16): Conv2d(256, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (17): ReLU(inplace)
          
          (18): MaxPool2d(kernel_size=2, stride=2, padding=0, dilation=1, ceil_mode=False)
          (19): Conv2d(256, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (20): ReLU(inplace)
          (21): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (22): ReLU(inplace)
          (23): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (24): ReLU(inplace)
          (25): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (26): ReLU(inplace)
          
          (27): MaxPool2d(kernel_size=2, stride=2, padding=0, dilation=1, ceil_mode=False)
          (28): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (29): ReLU(inplace)
          (30): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (31): ReLU(inplace)
          (32): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (33): ReLU(inplace)
          (34): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
          (35): ReLU(inplace)
          
          (36): MaxPool2d(kernel_size=2, stride=2, padding=0, dilation=1, ceil_mode=False)
        )
        """
        # hierarchy 1 (level 1)
        self.conv1_1 = features[0]
        self.relu1_1 = features[1]
        self.conv1_2 = features[2]
        self.relu1_2 = features[3]

        # hierarchy 2 (level 2)
        self.pool1 = features[4]
        self.conv2_1 = features[5]
        self.relu2_1 = features[6]
        self.conv2_2 = features[7]
        self.relu2_2 = features[8]

        # hierarchy 3 (level 3)
        self.pool2 = features[9]
        self.conv3_1 = features[10]
        self.relu3_1 = features[11]
        self.conv3_2 = features[12]
        self.relu3_2 = features[13]
        self.conv3_3 = features[14]
        self.relu3_3 = features[15]
        self.conv3_4 = features[16]
        self.relu3_4 = features[17]

        # hierarchy 4 (level 4)
        self.pool3 = features[18]
        self.conv4_1 = features[19]
        self.relu4_1 = features[20]
        self.conv4_2 = features[21]
        self.relu4_2 = features[22]
        self.conv4_3 = features[23]
        self.relu4_3 = features[24]
        self.conv4_4 = features[25]
        self.relu4_4 = features[26]

        # hierarchy 5 (level 5)
        self.pool4 = features[27]
        self.conv5_1 = features[28]
        self.relu5_1 = features[29]
        self.conv5_2 = features[30]
        self.relu5_2 = features[31]
        self.conv5_3 = features[32]
        self.relu5_3 = features[33]
        self.conv5_4 = features[34]
        self.relu5_4 = features[35]

        self.pool5 = features[36]

        # don't need the gradients, just want the features
        if not gradient:
            for param in self.parameters():
                param.requires_grad = False

        # reflection padding
        self.pad = nn.ReflectionPad2d(padding=1)

        # avg pooling
        if pool == 'avg':
            avg_pool = nn.AvgPool2d(kernel_size=2, stride=2)
            #         self.pool3 = self.pool4 = self.pool5 = avg_pool
            self.pool1 = self.pool2 = self.pool3 = self.pool4 = self.pool5 = avg_pool

    def forward(self, x, feature_layers):
        # level 1    pad before conv
        x = self.pad(x)
        conv1_1 = self.conv1_1(x)
        relu1_1 = self.relu1_1(conv1_1)
        conv1_2 = self.conv1_2(self.pad(relu1_1))
        relu1_2 = self.relu1_2(conv1_2)
        pool1 = self.pool1(relu1_2)

        # level 2
        pool1 = self.pad(pool1)
        conv2_1 = self.conv2_1(pool1)
        relu2_1 = self.relu2_1(conv2_1)
        conv2_2 = self.conv2_2(self.pad(relu2_1))
        relu2_2 = self.relu2_2(conv2_2)
        pool2 = self.pool2(relu2_2)

        # level 3
        pool2 = self.pad(pool2)
        conv3_1 = self.conv3_1(pool2)
        relu3_1 = self.relu3_1(conv3_1)
        conv3_2 = self.conv3_2(self.pad(relu3_1))
        relu3_2 = self.relu3_2(conv3_2)
        conv3_3 = self.conv3_3(self.pad(relu3_2))
        relu3_3 = self.relu3_3(conv3_3)
        conv3_4 = self.conv3_4(self.pad(relu3_3))
        relu3_4 = self.relu3_4(conv3_4)
        pool3 = self.pool3(relu3_4)

        # level 4
        pool3 = self.pad(pool3)
        conv4_1 = self.conv4_1(pool3)
        relu4_1 = self.relu4_1(conv4_1)
        conv4_2 = self.conv4_2(self.pad(relu4_1))
        relu4_2 = self.relu4_2(conv4_2)
        conv4_3 = self.conv4_3(self.pad(relu4_2))
        relu4_3 = self.relu4_3(conv4_3)
        conv4_4 = self.conv4_4(self.pad(relu4_3))
        relu4_4 = self.relu4_4(conv4_4)
        pool4 = self.pool4(relu4_4)

        # level 5
        pool4 = self.pad(pool4)
        conv5_1 = self.conv5_1(pool4)
        relu5_1 = self.relu5_1(conv5_1)
        conv5_2 = self.conv5_2(self.pad(relu5_1))
        relu5_2 = self.relu5_2(conv5_2)
        conv5_3 = self.conv5_3(self.pad(relu5_2))
        relu5_3 = self.relu5_3(conv5_3)
        conv5_4 = self.conv5_4(self.pad(relu5_3))
        relu5_4 = self.relu5_4(conv5_4)
        # pool5 = self.pool5(relu5_4)

        out = {
            'relu1_1': relu1_1,
            'relu1_2': relu1_2,

            'relu2_1': relu2_1,
            'relu2_2': relu2_2,

            'relu3_1': relu3_1,
            'relu3_2': relu3_2,
            'relu3_3': relu3_3,
            'relu3_4': relu3_4,

            'relu4_1': relu4_1,
            'relu4_2': relu4_2,
            'relu4_3': relu4_3,
            'relu4_4': relu4_4,

            'relu5_1': relu5_1,
            'relu5_2': relu5_2,
            'relu5_3': relu5_3,
            'relu5_4': relu5_4,


            #             'conv1_1': conv1_1,
            #             'conv1_2': conv1_2,

            #             'conv2_1': conv2_1,
            #             'conv2_2': conv2_2,

            #             'conv3_1': conv3_1,
            #             'conv3_2': conv3_2,
            #             'conv3_3': conv3_3,
            #             'conv3_4': conv3_4,

            #             'conv4_1': conv4_1,
            #             'conv4_2': conv4_2,
            #             'conv4_3': conv4_3,
            #             'conv4_4': conv4_4,

            #             'conv5_1': conv5_1,
            #             'conv5_2': conv5_2,
            #             'conv5_3': conv5_3,
            #             'conv5_4': conv5_4,
        }
        return dict((key, value) for key, value in out.items() if key in feature_layers)


if __name__ == '__main__':
    x = torch.randn((10, 5, 256, 256), dtype=torch.float32)
    model = VGG19()
    for i in range(100):
        print(i)
        y = model(x, feature_layers=("relu1_1", "relu1_2", "relu2_1", "relu2_2",
                                     "relu3_1", "relu3_2", "relu3_3", "relu3_4",
                                     "relu4_1", "relu4_2", "relu4_3", "relu4_4",
                                     "relu5_1", "relu4_2", "relu5_3", "relu5_4"))
        a = 3



import random
import cv2
import numpy as np


"""
    Implementation based on https://github.com/BadSeedX/SSAPS
    Paper: https://openaccess.thecvf.com/content/ACCV2022/html/Long_Self-Supervised_Augmented_Patches_Segmentation_for_Anomaly_Detection_ACCV_2022_paper.html
"""


class PatchAnomalyCreator(object):

    def __call__(self, img):
        origin_image = img.copy()
        origin_image = cv2.cvtColor(np.asarray(origin_image), cv2.COLOR_RGB2BGR)
        img_size = origin_image.shape[0]

        height, width, channel = origin_image.shape
        copy_raw = np.zeros((height, width, channel), np.uint8)  # mask image
        copy_image = origin_image.copy()

        # number of patch; size of patch; irregularity of patch
        outlines = []
        size = img_size // 6     # originally 163
        for i in range(4):
            outline = []
            border_value = img_size // 2.5      # originally 384
            (x_boarder, y_boarder) = np.random.randint(0, border_value - size, size=2)
            its = np.random.randint(4, 12)
            for it in range(its):
                x = np.random.randint(x_boarder, x_boarder + size)
                y = np.random.randint(y_boarder, y_boarder + size)
                outline.append([x, y])
            outlines.append(outline)
        direction = [[0, 0], [0, width // 2], [height // 2, 0], [height // 2, width // 2]]
        index = 0

        mask_images = []
        for outline in outlines:
            x_coords, y_coords = [item[1] for item in outline], [item[0] for item in outline]
            x_min, x_max, y_min, y_max = np.min(x_coords), np.max(x_coords), np.min(y_coords), np.max(y_coords)

            polygon = np.array(outline)
            copy = copy_raw.copy()
            cv2.fillPoly(copy, [polygon], (255, 255, 255))  # The polygon part turns white
            mask_index = list(np.where(copy > 0))
            copy[tuple(mask_index)] = origin_image[tuple(mask_index)]  # Add mask with I

            mask = copy[x_min:x_max, y_min:y_max]

            pads = self.__get_pads(height, width, abs(x_max - x_min), abs(y_max - y_min))
            up_pad, down_pad, left_pad, right_pad = pads[index][0], pads[index][1], pads[index][2], pads[index][3]

            mask_image = cv2.copyMakeBorder(mask, int(up_pad), int(down_pad), int(left_pad), int(right_pad),
                                            cv2.BORDER_CONSTANT, value=[0, 0, 0])
            mask_image = cv2.resize(mask_image, (height // 2, width // 2))
            mask_images.append(mask_image)  # Process mask polygon position

            x_coords_ = [int(x_coord + pads[index][0] - x_min) for x_coord in x_coords]
            y_coords_ = [int(y_coord + pads[index][2] - y_min) for y_coord in y_coords]
            outline_ = []
            for i in range(len(x_coords_)):
                outline_.append([y_coords_[i] + direction[index][1], x_coords_[i] + direction[index][0]])
            polygon_ = np.array(outline_)
            cv2.fillPoly(copy_image, [polygon_], (0, 0, 0))  # crop p in I

            index += 1

        top_half = cv2.hconcat([mask_images[0], mask_images[1]])
        bottom_half = cv2.hconcat([mask_images[2], mask_images[3]])
        result_image = cv2.vconcat([top_half, bottom_half])
        img_abnormal = cv2.add(result_image, copy_image)
        img_abnormal = cv2.cvtColor(img_abnormal, cv2.COLOR_BGR2RGB)

        gt_mic = copy_image.copy()
        gt_mic[gt_mic != 0] = 255
        white_index = list(np.where(gt_mic == 0))
        black_index = list(np.where(gt_mic == 255))
        gt_mic[tuple(white_index)] = 255
        gt_mic[tuple(black_index)] = 0

        gt_org = np.zeros_like(copy_image)

        gt_org = cv2.cvtColor(gt_org, cv2.COLOR_BGR2GRAY)
        gt_mic = cv2.cvtColor(gt_mic, cv2.COLOR_BGR2GRAY)
        gt_mic = gt_mic.astype('float32') / 255
        gt_org = gt_org.astype('float32') / 255

        mask_normal = cv2.resize(gt_org, (img_size, img_size)).astype(np.float32)
        mask_abnormal = cv2.resize(gt_mic, (img_size, img_size)).astype(np.float32)

        return np.array(img), img_abnormal, mask_normal, mask_abnormal

    def __get_pads(self, height, width, x, y):
        upper_bound, lower_bound = 0.5, 0.9
        lam1, lam2 = random.uniform(upper_bound, lower_bound), random.uniform(upper_bound, lower_bound)
        lam3, lam4 = random.uniform(upper_bound, lower_bound), random.uniform(upper_bound, lower_bound)
        lam5, lam6 = random.uniform(upper_bound, lower_bound), random.uniform(upper_bound, lower_bound)
        lam7, lam8 = random.uniform(upper_bound, lower_bound), random.uniform(upper_bound, lower_bound)
        pads = [[(height / 2 - x) * lam1, (height / 2 - x) * (1 - lam1), (width / 2 - y) * lam2,
                 (width / 2 - y) * (1 - lam2)],
                [(height / 2 - x) * lam3, (height / 2 - x) * (1 - lam3), (width / 2 - y) * (1 - lam4),
                 (width / 2 - y) * lam4],
                [(height / 2 - x) * (1 - lam5), (height / 2 - x) * lam5, (width / 2 - y) * lam6,
                 (width / 2 - y) * (1 - lam6)],
                [(height / 2 - x) * (1 - lam7), (height / 2 - x) * lam7, (width / 2 - y) * (1 - lam8),
                 (width / 2 - y) * lam8]]  # Up, Down, Left, Right
        return pads



import torch
import torch.nn.functional as F
import os
import numpy as np
import torchvision.transforms.functional as TF
import torchvision
from typing import List

from PIL import Image
from torchvision.transforms import transforms, Resize
from typing_extensions import Literal

from src.anoseg.models.DFC.backbone.vgg19 import VGG19
from src.anoseg.models.DFC.backbone.vgg19_s import VGG19_S
from src.anoseg.models.utils import BaseTester


class Tester(BaseTester):

    # region init

    _DATASET_TYPES = Literal["textures", "objects"]
    _CNN_LAYERS_TEXTURES = ("relu4_1", "relu4_2", "relu4_3", "relu4_4")
    _CNN_LAYERS_OBJECTS = ("relu4_3", "relu4_4", "relu5_1", "relu5_2")

    def __init__(self, model_path: str, dataset_type: _DATASET_TYPES, pretrained_weights_dir: str = None,
                 debugging: bool = False, patch_size: int = 256,
                 image_size: int = 1024, mask_size: int = 1024, use_self_ensembling: bool = False,
                 rot_90: bool = False, rot_180: bool = False, rot_270: bool = False, h_flip: bool = False,
                 h_flip_rot_90: bool = False, h_flip_rot_180: bool = False, h_flip_rot_270: bool = False,
                 integration_limit: float = 0.3):
        self.patch_size = patch_size
        self.resize_transform = Resize(size=patch_size, interpolation=TF.InterpolationMode.BILINEAR,
                                       antialias=True)

        super().__init__(model_path=model_path, debugging=debugging, image_size=image_size, mask_size=mask_size,
                         use_self_ensembling=use_self_ensembling, rot_90=rot_90, rot_180=rot_180, rot_270=rot_270,
                         h_flip=h_flip, h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                         h_flip_rot_270=h_flip_rot_270, integration_limit=integration_limit)

        self.dataset_type = dataset_type
        self.pretrained_weights_dir = pretrained_weights_dir

        self.__init_feat_layers()
        self.__load_models()

    def __load_models(self) -> None:
        big_model_path = os.path.join(self.model_path, 'big/match.pt')
        medium_model_path = os.path.join(self.model_path, 'medium/match.pt')
        small_model_path = os.path.join(self.model_path, 'small/match.pt')

        # pretrained feature extraction net
        self.feature_extraction = VGG19(pretrain=True, gradient=False, pool='avg',
                                        pretrained_weights_dir=self.pretrained_weights_dir).to(self.device)
        self.feature_extraction.eval()

        # trained feature estimation net
        self.feature_matching_big = VGG19_S(pretrain=False, gradient=True, pool='avg').to(self.device)
        self.feature_matching_big.load_state_dict(torch.load(big_model_path,
                                                             map_location=torch.device(self.device)))
        self.feature_matching_big.eval()

        self.feature_matching_medium = VGG19_S(pretrain=False, gradient=True, pool='avg').to(self.device)
        self.feature_matching_medium.load_state_dict(torch.load(medium_model_path,
                                                                map_location=torch.device(self.device)))
        self.feature_matching_medium.eval()

        self.feature_matching_small = VGG19_S(pretrain=False, gradient=True, pool='avg').to(self.device)
        self.feature_matching_small.load_state_dict(torch.load(small_model_path,
                                                               map_location=torch.device(self.device)))
        self.feature_matching_small.eval()

    def __init_feat_layers(self):
        cnn_layers_textures = ("relu4_1", "relu4_2", "relu4_3", "relu4_4")
        cnn_layers_objects = ("relu4_3", "relu4_4", "relu5_1", "relu5_2")
        if self.dataset_type == 'objects':
            self.feat_layers = cnn_layers_objects
        elif self.dataset_type == 'textures':
            self.feat_layers = cnn_layers_textures
        else:
            print('Unknown dataset type.')

    # endregion

    # region implement abstract methods

    def score_with_augmentation(self, img_input):
        score_list = self._get_self_ensembling_scores(img_input)
        final_score = self._combine_scores(score_list)

        return final_score

    def score(self, img_input) -> np.array:
        big_patches_score = self.__score_big_patches(img_input, 0)
        medium_patches_score = self.__score_medium_patches(img_input, 0)
        small_patches_score = self.__score_small_patches(img_input, 0)

        final_score = self.__calc_final_score(big_patches_score, medium_patches_score, small_patches_score)

        return final_score

    def preprocess_img(self, image_path: str, mean: List[float], std: List[float]):
        original = Image.open(image_path).convert('RGB')

        normalize = transforms.Normalize(mean=mean, std=std)
        resize = torchvision.transforms.Resize(size=self.image_size, interpolation=TF.InterpolationMode.BILINEAR,
                                               antialias=True)

        transform = transforms.Compose([transforms.ToTensor(), normalize, resize])
        preprocessed = transform(original)

        preprocessed = preprocessed[None, :]

        return preprocessed

    # endregion

    # region private methods

    # maximum of big, medium, small at each pixel
    def __calc_final_score(self, big, medium, small):
        # score = np.maximum(big, np.maximum(medium, small))
        score = (big + medium + small) / 3
        return score

    def __score_big_patches(self, img, threshold):
        return self.__score_patch(img, self.feature_matching_big, threshold, 1024)

    def __score_medium_patches(self, img, threshold):
        score = np.zeros(shape=(self.mask_size, self.mask_size), dtype=float)

        width, height = 512, 512

        for x in range(0, 1024, width):
            for y in range(0, 1024, height):
                patch = img[:, :, x:x + width, y:y + height]

                patch_score = self.__score_patch(patch, self.feature_matching_medium, threshold, 512)
                score_x = x
                score_y = y
                score[score_x:score_x + width, score_y:score_y + height] = patch_score

        return score

    def __score_small_patches(self, img, threshold):
        score = np.zeros(shape=(self.image_size, self.image_size), dtype=float)

        width, height = 256, 256

        for x in range(0, 1024, width):
            for y in range(0, 1024, height):
                patch = img[:, :, x:x + width, y:y + height]

                patch_score = self.__score_patch(patch, self.feature_matching_small, threshold, 256)
                score_x = x
                score_y = y
                score[score_x:score_x + width, score_y:score_y + height] = patch_score

        return score

    def __score_patch(self, patch, model, threshold, out_size):
        img = self.resize_transform(patch).to(self.device)

        with torch.no_grad():
            surrogate_label = self.feature_extraction(img, self.feat_layers)
            prediction = model(img, self.feat_layers)

        anomaly_map = 0
        for feat_layer in self.feat_layers:
            anomaly_map += F.interpolate(
                torch.pow(surrogate_label[feat_layer] - prediction[feat_layer], 2).mean(dim=1, keepdim=True),
                size=(out_size, out_size), mode="bilinear", align_corners=True)

        score = anomaly_map.data.cpu().numpy().squeeze()
        score[score < threshold] = 0

        return score

    # endregion



from typing import Tuple

from torch import nn, Tensor
from torchvision.models import resnet18, resnet50, wide_resnet50_2


class EncoderBase(nn.Module):
    def __init__(self, resnet_builder) -> None:
        super().__init__()
        self.resnet = resnet_builder(pretrained=True)

    def forward(self, x: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        """Return the three intermediary layers from the ResNet
        pre-trained model.
        Params
        ======
            x: Tensor - the input tensor of size (b * c * w * h)
        Returns
        =======
            feature_1: Tensor - the residual from layer 1
            feature_2: Tensor - the residual from layer 2
            feature_3: Tensor - the residual from layer 3
        """
        x = self.resnet.conv1(x)
        x = self.resnet.bn1(x)
        x = self.resnet.relu(x)
        x = self.resnet.maxpool(x)

        feature_1 = self.resnet.layer1(x)
        feature_2 = self.resnet.layer2(feature_1)
        feature_3 = self.resnet.layer3(feature_2)

        return feature_1, feature_2, feature_3


class ResNet50(EncoderBase):
    embeddings_size = 1792
    num_patches = 32 * 32

    def __init__(self) -> None:
        super(ResNet50, self).__init__(resnet50)


class ResNet18(EncoderBase):
    embeddings_size = 448
    num_patches = 32 * 32

    def __init__(self) -> None:
        super(ResNet18, self).__init__(resnet18)


class WideResNet50(EncoderBase):
    embeddings_size = 1792
    num_patches = 104 * 104

    def __init__(self) -> None:
        super(WideResNet50, self).__init__(wide_resnet50_2)



import math
import numpy as np
import torchvision
import torchvision.transforms.functional as TF
import torch.nn.functional as F
from os.path import join
from typing import List
from PIL import Image
from torchvision.transforms import transforms, Resize

from src.anoseg.models.PaDiM.backbone.padim import PaDiM
from src.anoseg.models.utils import BaseTester


class Tester(BaseTester):

    # region init

    def __init__(self, model_path: str, debugging: bool = False,
                 image_size: int = 1024, mask_size: int = 1024, patch_size: int = 256,
                 use_self_ensembling: bool = False,
                 rot_90: bool = False, rot_180: bool = False, rot_270: bool = False, h_flip: bool = False,
                 h_flip_rot_90: bool = False, h_flip_rot_180: bool = False, h_flip_rot_270: bool = False,
                 integration_limit: float = 0.3, backbone: str = "resnet18"):
        self.patch_size = patch_size
        self.resize_transform = Resize(size=patch_size, interpolation=TF.InterpolationMode.BILINEAR,
                                       antialias=True)

        super().__init__(model_path=model_path, debugging=debugging, image_size=image_size, mask_size=mask_size,
                         use_self_ensembling=use_self_ensembling, rot_90=rot_90, rot_180=rot_180, rot_270=rot_270,
                         h_flip=h_flip, h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                         h_flip_rot_270=h_flip_rot_270, integration_limit=integration_limit)
        self.backbone = backbone

        self.__load_model()

    def __load_model(self):
        self.padim_big = self.__load_patch_model(join(self.model_path, "big"))
        self.padim_medium = self.__load_patch_model(join(self.model_path, "medium"))
        self.padim_small = self.__load_patch_model(join(self.model_path, "small"))

    def __load_patch_model(self, model_path):
        n = np.load(join(model_path, "n.npy")).item()
        means = np.load(join(model_path, "means.npy"))
        covs = np.load(join(model_path, "covs.npy"))
        embedding_ids = np.load(join(model_path, "embedding_ids.npy"))
        padim = PaDiM.from_residuals(N=n,
                                     means=means,
                                     covs=covs,
                                     embedding_ids=embedding_ids,
                                     backbone=self.backbone,
                                     device=self.device,
                                     img_size=self.image_size)
        return padim

    # endregion

    # region implement abstract methods

    def score(self, img_input) -> np.array:
        big_patches_score = self.__score_big_patches(img_input)
        medium_patches_score = self.__score_medium_patches(img_input)
        small_patches_score = self.__score_small_patches(img_input)

        final_score = self.__calc_final_score(big_patches_score, medium_patches_score, small_patches_score)

        return final_score

    def score_with_augmentation(self, img_input) -> np.array:
        score_list = self._get_self_ensembling_scores(img_input)
        final_score = self._combine_scores(score_list)

        return final_score

    def preprocess_img(self, image_path: str, mean: List[float], std: List[float]):
        original = Image.open(image_path).convert('RGB')

        normalize = transforms.Normalize(mean=mean, std=std)
        resize = torchvision.transforms.Resize(size=self.patch_size, interpolation=TF.InterpolationMode.BILINEAR,
                                               antialias=True)

        transform = transforms.Compose([transforms.ToTensor(), normalize, resize])
        preprocessed = transform(original)

        preprocessed = preprocessed[None, :]

        return preprocessed

    # endregion

    # region private methods

    def __calc_final_score(self, big, medium, small) -> np.array:
        # score = np.max(big, np.max(medium, small))
        score = (big + medium + small) / 3
        return score

    def __score_big_patches(self, img) -> np.array:
        return self.__score_patch(img, self.padim_big, 1024)

    def __score_medium_patches(self, img) -> np.array:
        score = np.zeros(shape=(self.mask_size, self.mask_size), dtype=float)

        width, height = 512, 512

        for x in range(0, 1024, width):
            for y in range(0, 1024, height):
                patch = img[:, :, x:x + width, y:y + height]

                patch_score = self.__score_patch(patch, self.padim_medium, 512)
                score_x = x
                score_y = y
                score[score_x:score_x + width, score_y:score_y + height] = patch_score

        return score

    def __score_small_patches(self, img) -> np.array:
        score = np.zeros(shape=(self.mask_size, self.mask_size), dtype=float)

        width, height = 256, 256

        for x in range(0, 1024, width):
            for y in range(0, 1024, height):
                patch = img[:, :, x:x + width, y:y + height]

                patch_score = self.__score_patch(patch, self.padim_small, 256)
                score_x = x
                score_y = y
                score[score_x:score_x + width, score_y:score_y + height] = patch_score

        return score

    def __score_patch(self, patch, model, out_size) -> np.array:
        img = self.resize_transform(patch).to(self.device)

        distances = model.predict(img)
        w = int(math.sqrt(distances.numel()))
        raw_score = distances.reshape(1, 1, w, w)
        raw_score = F.interpolate(raw_score, size=(out_size, out_size), mode="bilinear",
                                  align_corners=True)

        raw_score = raw_score.detach().cpu().numpy().squeeze()

        return raw_score

    # end region



import torch
import torch.nn as nn

"""
    Code based on https://github.com/YoungGod/DFC
    Paper: https://doi.org/10.1016/j.patcog.2022.108874
"""


__all__ = [
    'VGG', 'vgg11', 'vgg11_bn', 'vgg13', 'vgg13_bn', 'vgg16', 'vgg16_bn',
    'vgg19_bn', 'vgg19',
]

model_urls = {
    'vgg11': 'https://download.pytorch.org/models/vgg11-bbd30ac9.pth',
    'vgg13': 'https://download.pytorch.org/models/vgg13-c768596a.pth',
    'vgg16': 'https://download.pytorch.org/models/vgg16-397923af.pth',
    'vgg19': 'https://download.pytorch.org/models/vgg19-dcbb9e9d.pth',
    'vgg11_bn': 'https://download.pytorch.org/models/vgg11_bn-6002323d.pth',
    'vgg13_bn': 'https://download.pytorch.org/models/vgg13_bn-abd245e5.pth',
    'vgg16_bn': 'https://download.pytorch.org/models/vgg16_bn-6c64b313.pth',
    'vgg19_bn': 'https://download.pytorch.org/models/vgg19_bn-c79401a0.pth',
}

model_dirs = {
    'vgg11': '/home/jovyan/work/model-checkpoints/vgg11-bbd30ac9.pth',
    'vgg13': '/home/jovyan/work/model-checkpoints/vgg13-c768596a.pth',
    'vgg16': '/home/jovyan/work/model-checkpoints/vgg16-397923af.pth',
    'vgg19': '/home/jovyan/work/model-checkpoints/vgg19-dcbb9e9d.pth',
}


class VGG(nn.Module):

    def __init__(self, features, num_classes=1000, init_weights=True):
        super(VGG, self).__init__()
        self.features = features
        self.avgpool = nn.AdaptiveAvgPool2d((7, 7))
        self.classifier = nn.Sequential(
            nn.Linear(512 * 7 * 7, 4096),
            nn.ReLU(True),
            nn.Dropout(),
            nn.Linear(4096, 4096),
            nn.ReLU(True),
            nn.Dropout(),
            nn.Linear(4096, num_classes),
        )
        if init_weights:
            self._initialize_weights()

    def forward(self, x):
        x = self.features(x)
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)


def make_layers(cfg, batch_norm=False):
    layers = []
    in_channels = 3
    for v in cfg:
        if v == 'M':
            layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
        else:
            # pad2d = models.ReflectionPad2d(padding=1)
            # layers += [pad2d]
            ## print("##################Reflection Padding#######################")
            conv2d = nn.Conv2d(in_channels, v, kernel_size=3)
            # conv2d = models.Conv2d(in_channels, v, kernel_size=3, padding=1)
            if batch_norm:
                layers += [conv2d, nn.BatchNorm2d(v), nn.ReLU(inplace=True)]
            else:
                layers += [conv2d, nn.ReLU(inplace=True)]
            in_channels = v
    return nn.Sequential(*layers)


cfgs = {
    'A': [64, 'M', 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'B': [64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'D': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512, 'M', 512, 512, 512, 'M'],
    'E': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M'],
}


def _vgg(arch, cfg, batch_norm, pretrained, progress, pretrained_weights_dir=None, **kwargs):
    if pretrained:
        kwargs['init_weights'] = False
    model = VGG(make_layers(cfgs[cfg], batch_norm=batch_norm), **kwargs)
    if pretrained:
        # TODO
        state_dict = load_state_dict_from_url(model_urls[arch],
                                              progress=progress, model_dir=pretrained_weights_dir)
        # state_dict = torch.load(model_dirs[arch])
        model.load_state_dict(state_dict)
    return model


def vgg11(pretrained=False, progress=True, **kwargs):
    r"""VGG 11-layer model (configuration "A") from
    `"Very Deep Convolutional Networks For Large-Scale Image Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _vgg('vgg11', 'A', False, pretrained, progress, **kwargs)


def vgg11_bn(pretrained=False, progress=True, **kwargs):
    r"""VGG 11-layer model (configuration "A") with batch normalization
    `"Very Deep Convolutional Networks For Large-Scale Image Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _vgg('vgg11_bn', 'A', True, pretrained, progress, **kwargs)


def vgg13(pretrained=False, progress=True, **kwargs):
    r"""VGG 13-layer model (configuration "B")
    `"Very Deep Convolutional Networks For Large-Scale Image Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _vgg('vgg13', 'B', False, pretrained, progress, **kwargs)


def vgg13_bn(pretrained=False, progress=True, **kwargs):
    r"""VGG 13-layer model (configuration "B") with batch normalization
    `"Very Deep Convolutional Networks For Large-Scale Image Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _vgg('vgg13_bn', 'B', True, pretrained, progress, **kwargs)


def vgg16(pretrained=False, progress=True, **kwargs):
    r"""VGG 16-layer model (configuration "D")
    `"Very Deep Convolutional Networks For Large-Scale Image Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _vgg('vgg16', 'D', False, pretrained, progress, **kwargs)


def vgg16_bn(pretrained=False, progress=True, **kwargs):
    r"""VGG 16-layer model (configuration "D") with batch normalization
    `"Very Deep Convolutional Networks For Large-Scale Image Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _vgg('vgg16_bn', 'D', True, pretrained, progress, **kwargs)


def vgg19(pretrained_weights_dir=None, pretrained=False, progress=True, **kwargs):
    r"""VGG 19-layer model (configuration "E")
    `"Very Deep Convolutional Networks For Large-Scale Image Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _vgg('vgg19', 'E', False, pretrained, progress, pretrained_weights_dir=pretrained_weights_dir, **kwargs)


def vgg19_bn(pretrained_weights_dir=None, pretrained=False, progress=True, **kwargs):
    r"""VGG 19-layer model (configuration 'E') with batch normalization
    `"Very Deep Convolutional Networks For Large-Scale Image Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _vgg('vgg19_bn', 'E', True, pretrained, progress, pretrained_weights_dir=pretrained_weights_dir, **kwargs)


##################################
#  Download pre-trained VGGs
##################################
import os
import sys
import zipfile
import warnings
import errno
from urllib.parse import urlparse
from urllib.request import urlopen

import tempfile
import hashlib
import shutil
import re

try:
    from tqdm import tqdm
except ImportError:
    # fake tqdm if it's not installed
    class tqdm(object):

        def __init__(self, total=None, disable=False,
                     unit=None, unit_scale=None, unit_divisor=None):
            self.total = total
            self.disable = disable
            self.n = 0
            # ignore unit, unit_scale, unit_divisor; they're just for real tqdm

        def update(self, n):
            if self.disable:
                return

            self.n += n
            if self.total is None:
                sys.stderr.write("\r{0:.1f} bytes".format(self.n))
            else:
                sys.stderr.write("\r{0:.1f}%".format(100 * self.n / float(self.total)))
            sys.stderr.flush()

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            if self.disable:
                return

            sys.stderr.write('\n')

HASH_REGEX = re.compile(r'-([a-f0-9]*)\.')
ENV_TORCH_HOME = 'TORCH_HOME'
ENV_XDG_CACHE_HOME = 'XDG_CACHE_HOME'
DEFAULT_CACHE_DIR = '~/.cache'


def _get_torch_home():
    torch_home = os.path.expanduser(
        os.getenv(ENV_TORCH_HOME,
                  os.path.join(os.getenv(ENV_XDG_CACHE_HOME, DEFAULT_CACHE_DIR), 'torch')))
    return torch_home


def download_url_to_file(url, dst, hash_prefix=None, progress=True):
    r"""Download object at the given URL to a local path.

    Args:
        url (string): URL of the object to download
        dst (string): Full path where object will be saved, e.g. `/tmp/temporary_file`
        hash_prefix (string, optional): If not None, the SHA256 downloaded file should start with `hash_prefix`.
            Default: None
        progress (bool, optional): whether or not to display a progress bar to stderr
            Default: True

    Example:
        >>> torch.hub.download_url_to_file('https://s3.amazonaws.com/pytorch/models/resnet18-5c106cde.pth', '/tmp/temporary_file')

    """
    file_size = None
    # We use a different API for python2 since urllib(2) doesn't recognize the CA
    # certificates in older Python
    u = urlopen(url)
    meta = u.info()
    if hasattr(meta, 'getheaders'):
        content_length = meta.getheaders("Content-Length")
    else:
        content_length = meta.get_all("Content-Length")
    if content_length is not None and len(content_length) > 0:
        file_size = int(content_length[0])

    # We deliberately save it in a temp file and move it after
    # download is complete. This prevents a local working checkpoint
    # being overriden by a broken download.
    dst = os.path.expanduser(dst)
    dst_dir = os.path.dirname(dst)
    f = tempfile.NamedTemporaryFile(delete=False, dir=dst_dir)

    try:
        if hash_prefix is not None:
            sha256 = hashlib.sha256()
        with tqdm(total=file_size, disable=not progress,
                  unit='B', unit_scale=True, unit_divisor=1024) as pbar:
            while True:
                buffer = u.read(8192)
                if len(buffer) == 0:
                    break
                f.write(buffer)
                if hash_prefix is not None:
                    sha256.update(buffer)
                pbar.update(len(buffer))

        f.close()
        if hash_prefix is not None:
            digest = sha256.hexdigest()
            if digest[:len(hash_prefix)] != hash_prefix:
                raise RuntimeError('invalid hash value (expected "{}", got "{}")'
                                   .format(hash_prefix, digest))
        shutil.move(f.name, dst)
    finally:
        f.close()
        if os.path.exists(f.name):
            os.remove(f.name)


def load_state_dict_from_url(url, model_dir=None, map_location=None, progress=True, check_hash=False):
    r"""Loads the Torch serialized object at the given URL.

    If downloaded file is a zip file, it will be automatically
    decompressed.

    If the object is already present in `model_dir`, it's deserialized and
    returned.
    The default value of `model_dir` is ``$TORCH_HOME/checkpoints`` where
    environment variable ``$TORCH_HOME`` defaults to ``$XDG_CACHE_HOME/torch``.
    ``$XDG_CACHE_HOME`` follows the X Design Group specification of the Linux
    filesytem layout, with a default value ``~/.cache`` if not set.

    Args:
        url (string): URL of the object to download
        model_dir (string, optional): directory in which to save the object
        map_location (optional): a function or a dict specifying how to remap storage locations (see torch.load)
        progress (bool, optional): whether or not to display a progress bar to stderr.
            Default: True
        check_hash(bool, optional): If True, the filename part of the URL should follow the naming convention
            ``filename-<sha256>.ext`` where ``<sha256>`` is the first eight or more
            digits of the SHA256 hash of the contents of the file. The hash is used to
            ensure unique names and to verify the contents of the file.
            Default: False

    Example:
        >>> state_dict = torch.hub.load_state_dict_from_url('https://s3.amazonaws.com/pytorch/models/resnet18-5c106cde.pth')

    """
    # Issue warning to move data if old env is set
    if os.getenv('TORCH_MODEL_ZOO'):
        warnings.warn('TORCH_MODEL_ZOO is deprecated, please use env TORCH_HOME instead')

    if model_dir is None:
        torch_home = _get_torch_home()
        model_dir = os.path.join(torch_home, 'checkpoints')

    try:
        os.makedirs(model_dir)
    except OSError as e:
        if e.errno == errno.EEXIST:
            # Directory already exists, ignore.
            pass
        else:
            # Unexpected OSError, re-raise.
            raise

    parts = urlparse(url)
    filename = os.path.basename(parts.path)
    cached_file = os.path.join(model_dir, filename)
    if not os.path.exists(cached_file):
        print('Downloading: "{}" to {}\n'.format(url, cached_file))
        hash_prefix = HASH_REGEX.search(filename).group(1) if check_hash else None
        download_url_to_file(url, cached_file, hash_prefix, progress=progress)
    else:
        print("Using weights in ", cached_file)

    # Note: extractall() defaults to overwrite file if exists. No need to clean up beforehand.
    #       We deliberately don't handle tarfile here since our legacy serialization format was in tar.
    #       E.g. resnet18-5c106cde.pth which is widely used.
    if zipfile.is_zipfile(cached_file):
        with zipfile.ZipFile(cached_file) as cached_zipfile:
            members = cached_zipfile.infolist()
            if len(members) != 1:
                raise RuntimeError('Only one file(not dir) is allowed in the zipfile')
            cached_zipfile.extractall(model_dir)
            extraced_name = members[0].filename
            cached_file = os.path.join(model_dir, extraced_name)

    return torch.load(cached_file, map_location=map_location)



from setuptools import setup, find_packages

VERSION = '0.0.3'
DESCRIPTION = 'Anomaly Segmentation'
LONG_DESCRIPTION = 'A collection of Anomaly Segmentation methods and an interface for easy use.'

setup(
    name="anoseg",
    version=VERSION,
    author="Hendrik Meininger",
    author_email="H.Meininger@outlook.de",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=['matplotlib',
                      'numpy',
                      'opencv-python',
                      'pandas',
                      'pillow',
                      'torch',
                      'scipy',
                      'scikit-learn',
                      'scikit-image',
                      'torchvision',
                      ],

    keywords=['python', 'anomaly', 'segmentation', 'anomaly segmentation', 'dfc', 'padim', 'spade'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ]
)



import numpy as np
from bisect import bisect

"""
    Code was adjusted from the official mvtec evaluation code.
    The original code is available at https://www.mvtec.com/company/research/datasets/mvtec-ad.
    Scientific Papers:
        https://link.springer.com/content/pdf/10.1007/s11263-020-01400-4.pdf
        https://openaccess.thecvf.com/content_CVPR_2019/papers/Bergmann_MVTec_AD_--_A_Comprehensive_Real-World_Dataset_for_Unsupervised_Anomaly_CVPR_2019_paper.pdf
"""


def calculate_auc(x_values, y_values, x_max=None):
    """
    This function calculates the definite integral of a curve given by
    x- and corresponding y-values. In contrast to, e.g., 'numpy.trapz()',
    this function allows to define an upper bound to the integration range by
    setting a value x_max.

    Points that do not have a finite x or y value will be ignored with a
    warning.

    Args:
        x_values: Samples from the domain of the function to integrate
          Need to be sorted in ascending order. May contain the same value
          multiple times. In that case, the order of the corresponding
          y values will affect the integration with the trapezoidal rule.
        y_values: Values of the function corresponding to x values.
        x_max: Upper limit of the integration. The y value at max_x will be
          determined by interpolating between its neighbors. Must not lie
          outside the range of x.

    Returns:
        Area under the curve.
    """

    x_values = np.asarray(x_values)
    y_values = np.asarray(y_values)
    finite_mask = np.logical_and(np.isfinite(x_values), np.isfinite(y_values))
    if not finite_mask.all():
        print("WARNING: Not all x and y values passed to trapezoid(...)"
              " are finite. Will continue with only the finite values.")
    x_values = x_values[finite_mask]
    y_values = y_values[finite_mask]

    # Introduce a correction term if max_x is not an element of x.
    correction = 0.
    if x_max is not None:
        if x_max not in x_values:
            # Get the insertion index that would keep x sorted after
            # np.insert(x, ins, x_max).
            ins = bisect(x_values, x_max)
            # x_max must be between the minimum and the maximum, so the
            # insertion_point cannot be zero or len(x).
            assert 0 < ins < len(x_values)

            # Calculate the correction term which is the integral between
            # the last x[ins-1] and x_max. Since we do not know the exact value
            # of y at x_max, we interpolate between y[ins] and y[ins-1].
            y_interp = y_values[ins - 1] + ((y_values[ins] - y_values[ins - 1]) *
                                            (x_max - x_values[ins - 1]) /
                                            (x_values[ins] - x_values[ins - 1]))
            correction = 0.5 * (y_interp + y_values[ins - 1]) * (x_max - x_values[ins - 1])

        # Cut off at x_max.
        mask = x_values <= x_max
        x_values = x_values[mask]
        y_values = y_values[mask]

    # Return area under the curve using the trapezoidal rule.
    return np.sum(0.5 * (y_values[1:] + y_values[:-1]) * (x_values[1:] - x_values[:-1])) + correction



import numpy as np
from typing import List
import matplotlib.pyplot as plt
import math


def display_images(img_list: List[np.array], titles=None, cols=2) -> None:
    if titles is not None and len(titles) != len(img_list):
        titles = None

    num_images = len(img_list)
    num_rows = math.ceil(num_images / cols)

    for i in range(len(img_list)):
        image = img_list[i]
        plt.subplot(num_rows, cols, i + 1)

        if titles is not None:
            plt.title(titles[i])

        plt.imshow(image)

    plt.show()


def save_images(save_path, img_list: List[np.array], titles=None, cols=2) -> None:
    if titles is not None and len(titles) != len(img_list):
        titles = None

    num_images = len(img_list)
    num_rows = math.ceil(num_images / cols)

    for i in range(len(img_list)):
        image = img_list[i]
        plt.subplot(num_rows, cols, i + 1)

        if titles is not None:
            plt.title(titles[i])

        plt.imshow(image)
        plt.axis('off')

    plt.savefig(save_path, bbox_inches='tight', pad_inches=0, dpi=400)




import time


class Timer(object):

    last_time = None
    start_time = None
    time_measures = []

    @staticmethod
    def start_timer():
        Timer.start_time = time.perf_counter()
        Timer.last_time = time.perf_counter()
        Timer.time_measures = []

    @staticmethod
    def log_time(task_name: str):
        current_time = time.perf_counter()
        task_time = current_time - Timer.last_time
        Timer.time_measures.append((task_name, round(task_time, 4)))
        Timer.last_time = current_time

    @staticmethod
    def print_task_times():
        current_time = time.perf_counter()
        complete_time = round(current_time - Timer.start_time, 4)

        hours = complete_time // 3600
        minutes = (complete_time % 3600) // 60
        remaining_seconds = (complete_time % 3600) % 60

        print(f"Complete execution time: \t {hours} hours, {minutes} minutes, and {remaining_seconds} seconds")

        for task_name, task_time in Timer.time_measures:
            hours = task_time // 3600
            minutes = (task_time % 3600) // 60
            remaining_seconds = (task_time % 3600) % 60

            print(f"{task_name}: \t {hours} hours, {minutes} minutes, and {remaining_seconds} seconds")





"""
    Script was adjusted from the official mvtec evaluation code.
    The original code is available at https://www.mvtec.com/company/research/datasets/mvtec-ad.
    Scientific Papers:
        https://link.springer.com/content/pdf/10.1007/s11263-020-01400-4.pdf
        https://openaccess.thecvf.com/content_CVPR_2019/papers/Bergmann_MVTec_AD_--_A_Comprehensive_Real-World_Dataset_for_Unsupervised_Anomaly_CVPR_2019_paper.pdf
"""

import numpy as np


def compute_classification_roc(
        anomaly_maps,
        scoring_function,
        ground_truth_labels):
    """Compute the ROC curve for anomaly classification on the image level.

    Args:
        anomaly_maps: List of anomaly maps (2D numpy arrays) that contain
          a real-valued anomaly score at each pixel.
        scoring_function: Function that turns anomaly maps into a single
          real valued anomaly score.

        ground_truth_labels: List of integers that indicate the ground truth
          class for each input image. 0 corresponds to an anomaly-free sample
          while a value != 0 indicates an anomalous sample.

    Returns:
        fprs: List of false positive rates.
        tprs: List of correspoding true positive rates.
    """
    assert len(anomaly_maps) == len(ground_truth_labels)

    # Compute the anomaly score for each anomaly map.
    anomaly_scores = map(scoring_function, anomaly_maps)
    num_scores = len(anomaly_maps)

    # Sort samples by anomaly score. Keep track of ground truth label.
    sorted_samples = \
        sorted(zip(anomaly_scores, ground_truth_labels), key=lambda x: x[0])

    # Compute the number of OK and NOK samples from the ground truth.
    ground_truth_labels_np = np.array(ground_truth_labels)
    num_nok = ground_truth_labels_np[ground_truth_labels_np != 0].size
    num_ok = ground_truth_labels_np[ground_truth_labels_np == 0].size

    # Initially, every NOK sample is correctly classified as anomalous
    # (tpr = 1.0), and every OK sample is incorrectly classified as anomalous
    # (fpr = 1.0).
    fprs = [1.0]
    tprs = [1.0]

    # Keep track of the current number of false and true positive predictions.
    num_fp = num_ok
    num_tp = num_nok

    # Compute new true and false positive rates when successively increasing
    # the threshold.
    next_score = None
    highest_accuracy = 0
    optimal_threshold = -1

    for i, (current_score, label) in enumerate(sorted_samples):
        labels = [i[1] for i in sorted_samples]

        correct = 0
        lower_labels = labels[:i]
        upper_labels = labels[i:]

        correct += lower_labels.count(0)
        correct += upper_labels.count(1)

        accuracy = correct / len(labels)
        if accuracy > highest_accuracy:
            highest_accuracy = accuracy
            last_scores = sorted_samples[i - 1][0] if i > 0 else current_score
            optimal_threshold = (last_scores + current_score) / 2

        if label == 0:
            num_fp -= 1
        else:
            num_tp -= 1

        if i < num_scores - 1:
            next_score = sorted_samples[i + 1][0]
        else:
            next_score = None  # end of list

        if (next_score != current_score) or (next_score is None):
            fprs.append(num_fp / num_ok)
            tprs.append(num_tp / num_nok)

    # Return (FPR, TPR) pairs in increasing order.
    fprs = fprs[::-1]
    tprs = tprs[::-1]

    # print("Highest accuracy: " + str(highest_accuracy))
    # print("Optimal threshold: " + str(optimal_threshold))

    return fprs, tprs, optimal_threshold, highest_accuracy
